#include "DisplayN18.h"

DisplayN18::DisplayN18() : resetPin(P0_20), backlightPin(P0_19), rsPin(P0_7), csPin(P0_2), spi(P0_21, P0_22, P1_15) {
    
    //buffer = new unsigned char[512];
    //memset(buffer, 0, 512);
    buffer = new unsigned char[320];
    memset(buffer, 0, 320);
    
    this->resetPin.write(false);
    this->backlightPin.write(true);
    this->rsPin.write(false);
    
    this->spi.format(8, 3);
    this->spi.frequency(15000000);
    
    this->initialize();
}

void DisplayN18::writeCommand(unsigned char command) {
    this->rsPin.write(false);
    
    this->csPin.write(false);
    
    this->spi.write(command);
    
    this->csPin.write(true);
}

void DisplayN18::writeData(unsigned char data) {
    this->writeData(&data, 1);
}

void DisplayN18::writeData(const unsigned char* data, unsigned int length) {
    this->rsPin.write(true);
    
    this->csPin.write(false);
    
    for (unsigned int i = 0; i < length; i++)
        // Turbo Time! Use the Burst SPI FastWrite! (from Erik Olieman -> http://developer.mbed.org/users/Sissors/code/BurstSPI/)
        this->spi.fastWrite(data[i]);
    //   this->spi.write(data[i]);
    
    spi.clearRX();
    
    this->csPin.write(true);
}

void DisplayN18::reset() {
    this->resetPin.write(false);
    wait_ms(300);
    
    this->resetPin.write(true);
    wait_ms(500);
}

void DisplayN18::initialize() {
    this->reset();

    // Sleep Out - Turn sleep mode off
    this->writeCommand(0x11);
    
    wait_ms(120);

    // SETPOWER
    this->writeCommand(0xB1);
    this->writeData(0x01); this->writeData(0x2C); this->writeData(0x2D);
    // SETDISP
    this->writeCommand(0xB2);
    this->writeData(0x01); this->writeData(0x2C); this->writeData(0x2D);
    // SETLUT
    this->writeCommand(0xB3);
    this->writeData(0x01); this->writeData(0x2C); this->writeData(0x2D);
    this->writeData(0x01); this->writeData(0x2C); this->writeData(0x2D);

    // SETCYC
    this->writeCommand(0xB4);
    this->writeData(0x07);

    // SETSTBA
    this->writeCommand(0xC0);
    this->writeData(0xA2); this->writeData(0x02); this->writeData(0x84);
    // SETPFUSE
    this->writeCommand(0xC1); this->writeData(0xC5);
    // What does this do?
    this->writeCommand(0xC2);
    this->writeData(0x0A); this->writeData(0x00);
    // What does this do?
    this->writeCommand(0xC3);
    this->writeData(0x8A); this->writeData(0x2A);
    // SetID
    this->writeCommand(0xC4);
    this->writeData(0x8A); this->writeData(0xEE);
    // What does this do?
    this->writeCommand(0xC5);
    this->writeData(0x0E);

    // Set Memory Access Control
    this->writeCommand(0x36);
    this->writeData(0xA8);

    // Read Internal Oscillator
    this->writeCommand(0xe0);
    this->writeData(0x0f); this->writeData(0x1a);
    this->writeData(0x0f); this->writeData(0x18);
    this->writeData(0x2f); this->writeData(0x28);
    this->writeData(0x20); this->writeData(0x22);
    this->writeData(0x1f); this->writeData(0x1b);
    this->writeData(0x23); this->writeData(0x37); this->writeData(0x00);
    this->writeData(0x07);
    this->writeData(0x02); this->writeData(0x10);

    // Read Power
    this->writeCommand(0xe1);
    this->writeData(0x0f); this->writeData(0x1b);
    this->writeData(0x0f); this->writeData(0x17);
    this->writeData(0x33); this->writeData(0x2c);
    this->writeData(0x29); this->writeData(0x2e);
    this->writeData(0x30); this->writeData(0x30);
    this->writeData(0x39); this->writeData(0x3f);
    this->writeData(0x00); this->writeData(0x07);
    this->writeData(0x03); this->writeData(0x10);

    // Set the initial column and page address sets (clipping area)
    this->writeCommand(0x2a);
    this->writeData(0x00); this->writeData(0x00);
    this->writeData(0x00); this->writeData(0x7f);
    this->writeCommand(0x2b);
    this->writeData(0x00); this->writeData(0x00);
    this->writeData(0x00); this->writeData(0x9f);

    // Read Source Option
    this->writeCommand(0xF0);
    this->writeData(0x01);
    this->writeCommand(0xF6);
    this->writeData(0x00);

    // Set the pixel format (16 bits per pixel)
    this->writeCommand(0x3A);
    this->writeData(0x05);

    // Turn on the display
    this->writeCommand(0x29);

    this->clear();
}

unsigned short *DisplayN18::getLineBuffer()
{
    // *** Tried buffering up all the commands however I think we need to toggle the CS pin or something
    // in order to execute the command... so we cannot just buffer and send everything at once...
    // Anyhoo.. ran out of time so we have to try this some other time...
    // Have to return a pointer on a 4 byte boundary I think!
    //return (unsigned short *)&buffer[12];
    // Just return the base buffer for now
    return (unsigned short *)buffer;
}
    
void DisplayN18::blitLine(unsigned char line)
{
    this->setClippingArea(0, line, 159, 0);
    this->writeCommand(0x2C);
    this->writeData(buffer, 160 * 1 * 2);
    
    // *** Tried buffering up all the commands however I think we need to toggle the CS pin or something
    // in order to execute the command... so we cannot just buffer and send everything at once...
    // Anyhoo.. ran out of time so we have to try this some other time...
    /*
    // Setup column address set
    buffer[0] = 0;
    buffer[1] = 0x2A;
    buffer[2] = 0;
    buffer[3] = 0;      // x
    buffer[4] = 0;
    buffer[5] = 159;    // x + (width - 1)
    buffer[6] = 0x2B;
    buffer[7] = 0;
    buffer[8] = line;   // y
    buffer[9] = 0;
    buffer[10] = line;   // y + (height - 1)
    // Write Command
    buffer[11] = 0x2C;
    
    //buffer[0] = 0;
    //buffer[1] = 0x2C;
    // Everything else is a buffer of 160 unsigned shorts
    // Already should be filled out....
    
    // Write the buffer to the display
    writeData(&buffer[1], 11 + (160 * 2));
    */
}

void DisplayN18::setClippingArea(unsigned char x, unsigned char y, unsigned char width, unsigned char height) {
    unsigned char data[4] = { 0x00, 0x00, 0x00, 0x00 };

    // Set the Column Address Set
    data[1] = x;
    data[3] = x + width;
    this->writeCommand(0x2A);
    this->writeData(data, 4);

    // Set the Page Address Set
    data[1] = y;
    data[3] = y + height;
    this->writeCommand(0x2B);
    this->writeData(data, 4);
}

unsigned short DisplayN18::rgbToShort(unsigned char r, unsigned char g, unsigned char b) {   
unsigned short red = r;
    unsigned short green = g;
    unsigned short blue = b;

    red /= 8;
    green /= 4;
    blue /= 8;

    red &= 0x1F;
    green &= 0x3F;
    blue &= 0x1F;

    red <<= 8;//3;
    blue <<= 3;//8;
    green = ((green & 0x7) << 13) + ((green & 0x38) >> 3);

    return red | green | blue;
}

void DisplayN18::clear(unsigned short backColor) {
    for (unsigned int i = 0; i < LCD_WIDTH; i += 10)
        for (unsigned int j = 0; j < LCD_HEIGHT; j += 8)
            this->fillRect(i, j, 10, 8, backColor);
}

void DisplayN18::draw(const unsigned short* data, int x, int y, int width, int height) {
    this->setClippingArea(x, y, width - 1, height - 1);
    this->writeCommand(0x2C);
    this->writeData(reinterpret_cast<const unsigned char*>(data), width * height * 2);
}

void DisplayN18::setPixel(int x, int y, unsigned short foreColor) {
    this->draw(&foreColor, x, y, 1, 1);
}

void DisplayN18::fillRect(int x, int y, int width, int height, unsigned short foreColor) {
    this->setClippingArea(static_cast<unsigned char>(x), static_cast<unsigned char>(y), static_cast<unsigned char>(width - 1), static_cast<unsigned char>(height));
    
    this->writeCommand(0x2C);
    
    unsigned short buffer[50];
    for (int j = 0; j < sizeof(buffer) / 2; j++)
        buffer[j] = foreColor;

    this->rsPin.write(true);

    int i;
    for (i = sizeof(buffer); i < height * width * 2; i += sizeof(buffer))
        this->writeData(reinterpret_cast<unsigned char*>(buffer), sizeof(buffer));
    
    i -= sizeof(buffer);
    if (i != height * width * 2)
        this->writeData(reinterpret_cast<unsigned char*>(buffer), height * width * 2 - i);
}

void DisplayN18::drawRect(int x, int y, int width, int height, unsigned short foreColor) {
    this->drawLine(x, y, x + width, y, foreColor);
    this->drawLine(x, y + height, x + width, y + height, foreColor);
    this->drawLine(x, y, x, y + height, foreColor);
    this->drawLine(x + width, y, x + width, y + height, foreColor);
}

void DisplayN18::fillCircle(int x, int y, int radius, unsigned short foreColor) {
    int f = 1 - radius;
    int dd_f_x = 1;
    int dd_f_y = -2 * radius;
    int x1 = 0;
    int y1 = radius;

    for (int i = y - radius; i <= y + radius; i++)
        this->setPixel(x, i, foreColor);

    while (x1 < y1) {
        if (f >= 0) {
            y1--;
            dd_f_y += 2;
            f += dd_f_y;
        }

        x1++;
        dd_f_x += 2;
        f += dd_f_x;

        for (int i = y - y1; i <= y + y1; i++) {
            this->setPixel(x + x1, i, foreColor);
            this->setPixel(x - x1, i, foreColor);
        }

        for (int i = y - x1; i <= y + x1; i++) {
            this->setPixel(x + y1, i, foreColor);
            this->setPixel(x - y1, i, foreColor);
        }
    }
}

void DisplayN18::drawCircle(int x, int y, int radius, unsigned short foreColor) {
    int f = 1 - radius;
    int dd_f_x = 1;
    int dd_f_y = -2 * radius;
    int x1 = 0;
    int y1 = radius;

    this->setPixel(x, y + radius, foreColor);
    this->setPixel(x, y - radius, foreColor);
    this->setPixel(x + radius, y, foreColor);
    this->setPixel(x - radius, y, foreColor);

    while (x1 < y1) {
        if (f >= 0) {
            y1--;
            dd_f_y += 2;
            f += dd_f_y;
        }

        x1++;
        dd_f_x += 2;
        f += dd_f_x;

        this->setPixel(x + x1, y + y1, foreColor);
        this->setPixel(x - x1, y + y1, foreColor);
        this->setPixel(x + x1, y - y1, foreColor);
        this->setPixel(x - x1, y - y1, foreColor);

        this->setPixel(x + y1, y + x1, foreColor);
        this->setPixel(x - y1, y + x1, foreColor);
        this->setPixel(x + y1, y - x1, foreColor);
        this->setPixel(x - y1, y - x1, foreColor);
    }
}

void DisplayN18::drawLine(int x0, int y0, int x1, int y1, unsigned short foreColor) {
    if (x0 == x1) {
        if (y1 < y0) {
            int temp = y0;
            y0 = y1;
            y1 = temp;
        }

        this->setClippingArea(static_cast<unsigned char>(x0), static_cast<unsigned char>(y0), 0, static_cast<unsigned char>(y1 - y0 - 1));
        this->writeCommand(0x2C);

        unsigned short data[DisplayN18::STEP];
        for (int i = 0; i < DisplayN18::STEP; i++)
            data[i] = foreColor;

        for (unsigned char thisY = y0; thisY < y1; thisY += DisplayN18::STEP)
            this->writeData(reinterpret_cast<unsigned char*>(data), (thisY + DisplayN18::STEP <= y1 ? DisplayN18::STEP : y1 - thisY) * 2);

        return;
    }

    if (y0 == y1) {
        if (x1 < x0) {
            int temp = x0;
            x0 = x1;
            x1 = temp;
        }

        this->setClippingArea(static_cast<unsigned char>(x0), static_cast<unsigned char>(y0), static_cast<unsigned char>(x1 - x0 - 1), 0);
        this->writeCommand(0x2C);

        unsigned short data[DisplayN18::STEP];
        for (int i = 0; i < DisplayN18::STEP; i++)
            data[i] = foreColor;

        for (unsigned char thisX = x0; thisX < x1; thisX += DisplayN18::STEP)
            this->writeData(reinterpret_cast<unsigned char*>(data), (thisX + DisplayN18::STEP <= x1 ? DisplayN18::STEP : x1 - thisX) * 2);

        return;
    }

    int t;
    bool steep = ((y1 - y0) < 0 ? -(y1 - y0) : (y1 - y0)) > ((x1 - x0) < 0 ? -(x1 - x0) : (x1 - x0));

    if (steep) {
        t = x0;
        x0 = y0;
        y0 = t;
        t = x1;
        x1 = y1;
        y1 = t;
    }

    if (x0 > x1) {
        t = x0;
        x0 = x1;
        x1 = t;

        t = y0;
        y0 = y1;
        y1 = t;
    }

    int dx, dy;
    dx = x1 - x0;
    dy = (y1 - y0) < 0 ? -(y1 - y0) : (y1 - y0);

    int err = (dx / 2);
    int ystep;

    ystep = y0 < y1 ? 1 : -1;

    for (; x0 < x1; x0++) {
        if (steep)
            this->setPixel(y0, x0, foreColor);
        else
            this->setPixel(x0, y0, foreColor);

        err -= dy;

        if (err < 0) {
            y0 += (char)ystep;
            err += dx;
        }
    }
}
  
void DisplayN18::drawBitmap(int x, int y, int w, int h, unsigned char *bitmap, unsigned short foreColor, unsigned short backColor)
{
    this->setClippingArea(x, y, w - 1, h);
    this->writeCommand(0x2C);
    
    this->rsPin.write(true);
    this->csPin.write(false);
    
    int length = ((h * w) / 8);
    
    for (unsigned int i = 0; i < length; i++)
    {
        unsigned short color = 0;
        unsigned char data = bitmap[i];
        for (unsigned int bit = 0; bit < 8; bit++)
        {
            color = (data & (1 << bit)) ? foreColor : backColor;
            this->spi.write((unsigned char)(color >> 8));
            this->spi.write((unsigned char)(color & 0x00FF));
        }
    }
    
    this->csPin.write(true);
}
/*
void DisplayN18::drawCharacter(int x, int y, const char character, unsigned short foreColor, unsigned short backColor, unsigned char fontSize)
{
    if (character > 126 || character < 32) return;
    drawBitmap(x, y, CHAR_WIDTH, CHAR_HEIGHT, &font8x8_basic[(character - 32) * 8], foreColor, backColor);
}

void DisplayN18::drawString(int x, int y, const char* str, unsigned short foreColor, unsigned short backColor, unsigned char fontSize) {
    if (*str == '\0')
        return;

    do {
        this->drawCharacter(x, y, *str, foreColor, backColor, fontSize);
        
        x += (CHAR_WIDTH + CHAR_SPACING) * fontSize;
    } while (*(++str) != '\0');
}

unsigned char *DisplayN18::getCharBuffer()
{
    return font8x8_basic;
}

*/
