#include <stdio.h>
#include <string.h>
//#include "DisplayN18.h"
#include "LCD_ST7735.h"     // Retro LCD Library by Chris Taylor
#include "font8x8.h"        // 8x8 Font

#define CHAR_WIDTH      FONT8X8_WIDTH
#define CHAR_HEIGHT     FONT8X8_HEIGHT
#define CHAR_MAP_WIDTH  (LCD_WIDTH / CHAR_WIDTH)
#define CHAR_MAP_HEIGHT (LCD_HEIGHT / CHAR_HEIGHT)
#define CHAR_MAP_LEN    (CHAR_MAP_WIDTH * CHAR_MAP_HEIGHT)
#define CHAR_MAP_LAST_LINE (CHAR_MAP_LEN - CHAR_MAP_WIDTH)
#define COLOR_PAL_LEN   16
#define LINE_BUFFER_LEN (LCD_WIDTH * sizeof(uint16_t))

enum SHIFT_DIRECTION
{
    eShiftUp,
    eShiftDown,
    eShiftLeft,
    eShiftRight,
};

/*
#define PAL_TRANSPARENT         0
#define PAL_BLACK               1
#define PAL_MEDGREEN            2
#define PAL_LIGHTGREEN          3
#define PAL_DARKBLUE            4
#define PAL_LIGHTBLUE           5
#define PAL_DARKRED             6
#define PAL_CYAN                7
#define PAL_MEDRED              8
#define PAL_LIGHTRED            9
#define PAL_DARKYELLOW          10
#define PAL_LIGHTYELLOW         11
#define PAL_DARKGREEN           12
#define PAL_MAGENTA             13
#define PAL_GREY                14
#define PAL_WHITE               15
*/

#define PAL_WHITE               0
#define PAL_YELLOW              1
#define PAL_ORANGE              2
#define PAL_RED                 3
#define PAL_MAGENTA             4
#define PAL_DARKBLUE            5
#define PAL_BLUE                6
#define PAL_LIGHTBLUE           7
#define PAL_GREEN               8
#define PAL_DARKGREEN           9
#define PAL_BROWN               10
#define PAL_LIGHTBROWN          11
#define PAL_LIGHTGREY           12
#define PAL_GREY                13
#define PAL_BLACK               14
#define PAL_TRANSPARENT         15
#define PAL_DEFAULT             255

struct SPRITE
{
    bool enabled;
    uint8_t char_index;
    uint8_t x;
    uint8_t y;
    uint8_t fore_color;
    uint8_t back_color;
    bool midhandle;
};

#define MAX_SPRITES             64
#define MAX_USER_CHAR           64
#define MAX_USER_CHAR_BUF_LEN   (MAX_USER_CHAR * 8)

//class display : public DisplayN18
class display : public LCD_ST7735
{
public:

    display();
    ~display();

    void render();
    
    // Character Map Functions
    void shift_map(SHIFT_DIRECTION direction, int inc = 1, bool clear_after_shift = false);
    uint16_t get_map_info(uint8_t x, uint8_t y);
    uint8_t get_map_color(uint8_t x, uint8_t y);
    uint8_t get_map(uint8_t x, uint8_t y);
    void set_map(uint8_t char_index, uint8_t x, uint8_t y, uint8_t fore_color, uint8_t back_color);
    void set_map_info(uint16_t tile_info, uint8_t x, uint8_t y);
    void setcharat(uint8_t x, uint8_t y, uint8_t character, uint8_t fore_color = PAL_DEFAULT, uint8_t back_color = PAL_DEFAULT);
    
    // Character based printing
    void clear();
    void crlf();
    void setforecolor(uint8_t color) { m_fore_color = color; };
    void setbackcolor(uint8_t color) { m_back_color = color; };
    void setwordwrap(bool enable) { m_word_wrap = enable; };
    
    void print(char *pString);
    void print(int iNumber);
    void print(float fNumber);
    void print(const char *format, ...);
    
    void println(char *pString);
    void println(int iNumber);
    void println(float fNumber);
    void println(const char *format, ...);
    
    void printat(uint8_t x, uint8_t y, char *pString);
    void printat(uint8_t x, uint8_t y, int iNumber);
    void printat(uint8_t x, uint8_t y, float fNumber);
    void printat(uint8_t x, uint8_t y, const char *format, ...);
    
    // Sprite functions
    void AddSprite(int iSpriteNum, int iChar, int x, int y, int fore_color, int back_color, bool midhandle = false);  // width/height is a multiple of 8
    void EnableSprite(int iSpriteNum, bool bEnable);
    void RemoveSprite(int iSpriteNum);
    void SetSpriteChar(int iSpriteNum, int iChar);
    void SetSpritePos(int iSpriteNum, int x, int y);
    void SetSpriteColor(int iSpriteNum, int fore_color, int back_color);
    void SetMidHandle(int iSpriteNum, bool bEnable);
    bool SpriteCollision(int iSpriteNum1, int iSpriteNum2);
    
    void SetCustomChar(int iCharNum, uint8_t *pBuf, int iNum = 1, bool isMSX = false);         // Custom Characters can be used for Sprites and Tiles
    
protected:

    bool m_word_wrap;
    uint8_t m_fore_color, m_back_color;     // Indecies into the color palette
    uint8_t m_cursor_x, m_cursor_y;
    uint8_t m_char_map[CHAR_MAP_LEN];
    uint8_t m_color_map[CHAR_MAP_LEN];
    uint16_t m_color_pal[COLOR_PAL_LEN];
    uint8_t *m_line_buffer;
    
    void blit_line(uint8_t line);
    void bounds_check_and_scroll();
    int get_linear_cursor_pos() { return (m_cursor_x + (m_cursor_y * CHAR_MAP_WIDTH)); };
    int get_linear_pos(int x, int y) { return (x + (y * CHAR_MAP_WIDTH)); };
    
    SPRITE sprite_list[MAX_SPRITES];
    
    uint8_t user_char[MAX_USER_CHAR_BUF_LEN];
    void SetCustomChar8x8Norm(int iCharNum, uint8_t *pBuf);
    void SetCustomChar8x8MSX(int iCharNum, uint8_t *pBuf);
};

